/*
 * the following functions will add Solaris Audio support for 
 * RtAudio (http://www.music.mcgill.ca/~gary/rtaudio/)
 *
 * Joakim L. Gilje
 */
#if defined(__SOLARIS_AUDIO__)

#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/stat.h>

#include <fcntl.h>
#include <errno.h>
#include <unistd.h>

#define DAC_NAME "/dev/sound/"
#define MAX_DEVICES 16
#define MAX_CHANNELS 16

extern "C" void *solarisCallbackHandler(void * ptr);

RtApiSolaris :: RtApiSolaris()
{
  this->initialize();

  if (nDevices_ <= 0) {
    sprintf(message_, "RtApiSolaris: no Solaris audio devices found!");
    error(RtError::NO_DEVICES_FOUND);
 }
}

RtApiSolaris :: ~RtApiSolaris()
{
  if ( stream_.mode != UNINITIALIZED )
    closeStream();
}

void RtApiSolaris :: initialize(void)
{
  // Count cards and devices
  nDevices_ = 0;

  // This is really the bruteforce method from LINUX_OSS, searching in /dev/sound/(int)
  char device_name[16];
  int fd = 0;
  RtApiDevice device;
  for (int i=-1; i<MAX_DEVICES; i++) {

    // Probe /dev/audio first, since it is supposed to be the default device.
    if (i == -1)
      sprintf(device_name, "/dev/audio");
    else
      sprintf(device_name, "%s%d", DAC_NAME, i);

    // First try to open the device for playback, then record mode.
    fd = open(device_name, O_WRONLY | O_NONBLOCK);
    if (fd == -1) {
      // Open device for playback failed ... either busy or doesn't exist.
      if (errno != EBUSY && errno != EAGAIN) {
        // Try to open for capture
        fd = open(device_name, O_RDONLY | O_NONBLOCK);
        if (fd == -1) {
          // Open device for record failed.
          if (errno != EBUSY && errno != EAGAIN)
            continue;
          else {
            sprintf(message_, "RtApiSolairs: Solaris record device (%s) is busy.", device_name);
            error(RtError::WARNING);
            // still count it for now
          }
        }
      }
      else {
        sprintf(message_, "RtApiSolaris: Solaris playback device (%s) is busy.", device_name);
        error(RtError::WARNING);
        // still count it for now
      }
    }

    if (fd >= 0) close(fd);
    device.name.erase();
    device.name.append( (const char *)device_name, strlen(device_name)+1);
    devices_.push_back(device);
    nDevices_++;
  }
}

void RtApiSolaris :: probeDeviceInfo(RtApiDevice *info)
{
    // Probe Playback
    device_fd = open(info->name.c_str(), O_WRONLY);
    if (device_fd == -1) {
	sprintf(message_, "RtApiSolaris: %s failed to open for playback.", info->name.c_str());
	error(RtError::DEBUG_WARNING);
    } else {
	audio_device_t audio_dev;
	audio_info_t audio_info;
	int err, i;
	
	err = ioctl(device_fd, AUDIO_GETDEV, &audio_dev);
	if (err < 0) {
	    sprintf(message_, "RtApiSolaris: Failed to query AUDIO_GETDEV on %s.", info->name.c_str());
	    error(RtError::DEBUG_WARNING);
	}/* else {
	    printf("name: %s\nversion: %s\nconfig: %s\n", audio_dev.name, audio_dev.version, audio_dev.config);
	}*/
	
	AUDIO_INITINFO(&audio_info);
	info->minOutputChannels = MAX_CHANNELS;
	for (i = 1; i <= MAX_CHANNELS; i++) {
	    audio_info.play.channels = i;
	    err = ioctl(device_fd, AUDIO_SETINFO, &audio_info);
	    if (err == 0) {
		if (i < info->minOutputChannels) {
		    info->minOutputChannels = i;
		}
		
		info->maxOutputChannels = i;
	    }
	}
	
	close(device_fd);
    }

    // Probe Recording
    device_fd = open(info->name.c_str(), O_RDONLY);
    if (device_fd == -1) {
	sprintf(message_, "RtApiSolaris: %s failed to open for recording.", info->name.c_str());
	error(RtError::DEBUG_WARNING);
    } else {
	audio_info_t audio_info;
	int err, i;
	
	AUDIO_INITINFO(&audio_info);
	info->minInputChannels = MAX_CHANNELS;
	for (i = 1; i <= MAX_CHANNELS; i++) {
	    audio_info.record.channels = i;
	    err = ioctl(device_fd, AUDIO_SETINFO, &audio_info);
	    if (err == 0) {
		if (i < info->minInputChannels) {
		    info->minInputChannels = i;
		}
		
		info->maxInputChannels = i;
	    }
	}
	
	close(device_fd);
    }
    
    // Probe Duplex
    device_fd = open(info->name.c_str(), O_RDWR);
    if (device_fd == -1) {
	sprintf(message_, "RtApiSolaris: %s failed to open for duplex probing.", info->name.c_str());
	error(RtError::DEBUG_WARNING);
    } else {
	audio_info_t audio_info;
	int err, i;
	
	err = ioctl(device_fd, AUDIO_GETINFO, &audio_info);
	if ((audio_info.hw_features & AUDIO_HWFEATURE_DUPLEX) > 0) {
	    info->hasDuplexSupport = true;
	    info->maxDuplexChannels = info->maxOutputChannels;
	    info->minDuplexChannels = info->minOutputChannels;
	}
	
	close (device_fd);
    }
    
    // Probe bitwidths
    device_fd = open(info->name.c_str(), O_WRONLY);
    if (device_fd == -1) {
	sprintf(message_, "RtApiSolaris: %s failed to open for playback (bitwidth probe).", info->name.c_str());
	error(RtError::DEBUG_WARNING);
    } else {
	audio_info_t audio_info;
	int err, i;
	
	AUDIO_INITINFO(&audio_info);
	audio_info.play.encoding = AUDIO_ENCODING_LINEAR;
	audio_info.record.encoding = AUDIO_ENCODING_LINEAR;
	info->nativeFormats = 0;
	
	for (i = 8; i <= 32; i += 8) {
	    audio_info.play.precision = i;
	    audio_info.record.precision = i;
	    err = ioctl(device_fd, AUDIO_SETINFO, &audio_info);
	    if (err == 0) {
		switch (i) {
		    case 8:
			info->nativeFormats |= RTAUDIO_SINT8;
			break;
		    case 16:
			info->nativeFormats |= RTAUDIO_SINT16;
			break;
		    case 24:
			info->nativeFormats |= RTAUDIO_SINT24;
			break;
		    case 32:
			info->nativeFormats |= RTAUDIO_SINT32;
			break;
		}
	    }
	}
	
	if (info->nativeFormats == 0) {
            sprintf(message_, "RtApiSolaris: data format not supported by RtAudio.");
	    error(RtError::DEBUG_WARNING);
	}
	
	close(device_fd);
    }

    // Probe samplerates
    device_fd = open(info->name.c_str(), O_WRONLY);
    if (device_fd == -1) {
	sprintf(message_, "RtApiSolaris: %s failed to open for playback (samplerate probe).", info->name.c_str());
	error(RtError::DEBUG_WARNING);
    } else {
	audio_info_t audio_info;
	int err, i;
	
	AUDIO_INITINFO(&audio_info);
	for (i = 0; i <= MAX_SAMPLE_RATES; i++) {
	    audio_info.play.sample_rate = SAMPLE_RATES[i];
	    audio_info.record.sample_rate = SAMPLE_RATES[i];
	    err = ioctl(device_fd, AUDIO_SETINFO, &audio_info);
	    if (err == 0) {
		info->sampleRates.push_back(SAMPLE_RATES[i]);
	    }
	}
	
	close(device_fd);
    }
    
    info->probed = true;
    return;
}

bool RtApiSolaris :: probeDeviceOpen(int device, StreamMode mode, int channels, 
                                int sampleRate, RtAudioFormat format,
                                int *bufferSize, int numberOfBuffers)
{
    int err, i, precision;
    audio_info_t audio_info;
    const char *name = devices_[device].name.c_str();
    int *handle = (int *) stream_.apiHandle;
    
    if (mode == OUTPUT) {
	device_fd = open(name, O_WRONLY);
    } else if (mode == INPUT) {
	device_fd = open(name, O_RDONLY);
    } else if (mode == DUPLEX) {
	device_fd = open(name, O_RDWR);
    }
    
    if (device_fd < 0) {
	sprintf(message_, "RtApiSolaris: device failed to open (%s).", name);
	error(RtError::DEBUG_WARNING);
	
	return FAILURE;
    }
    
    switch (format) {
	case RTAUDIO_SINT8:
            if (devices_[device].nativeFormats & RTAUDIO_SINT8) {
                precision = 8;
                stream_.deviceFormat[mode] = RTAUDIO_SINT8;
            } else {
                precision = 16;
                stream_.deviceFormat[mode] = RTAUDIO_SINT16;
            }
	    break;
	case RTAUDIO_SINT16:
	    precision = 16;
	    stream_.deviceFormat[mode] = RTAUDIO_SINT16;
	    break;
	case RTAUDIO_SINT24:
            if (devices_[device].nativeFormats & RTAUDIO_SINT24) {
                precision = 24;
                stream_.deviceFormat[mode] = RTAUDIO_SINT24;
            } else {
                precision = 16;
                stream_.deviceFormat[mode] = RTAUDIO_SINT16;
            }
	    break;
	case RTAUDIO_SINT32:
            if (devices_[device].nativeFormats & RTAUDIO_SINT32) {
                precision = 32;
                stream_.deviceFormat[mode] = RTAUDIO_SINT32;
            } else {
                precision = 16;
                stream_.deviceFormat[mode] = RTAUDIO_SINT16;
            }
	    break;
        // not supported anyway
        case RTAUDIO_FLOAT32:
	case RTAUDIO_FLOAT64:
	    precision = 16;
	    stream_.deviceFormat[mode] = RTAUDIO_SINT16;
	    break;
    }
    stream_.userFormat = format;
    
    AUDIO_INITINFO(&audio_info);
    audio_info.play.sample_rate = sampleRate;
    audio_info.play.channels = channels;
    audio_info.play.precision = precision;
    audio_info.play.encoding = AUDIO_ENCODING_LINEAR;
    audio_info.play.buffer_size = *bufferSize * channels * 
                                   formatBytes(stream_.deviceFormat[mode]);
    audio_info.record.sample_rate = sampleRate;
    audio_info.record.channels = channels;
    audio_info.record.precision = precision;
    audio_info.record.encoding = AUDIO_ENCODING_LINEAR;
    audio_info.record.buffer_size = *bufferSize * channels * 
                                     formatBytes(stream_.deviceFormat[mode]);
    
    err = ioctl(device_fd, AUDIO_SETINFO, &audio_info);
    if (err != 0) {
	sprintf(message_, "RtApiSolaris: failed to set parameters in probeOpenDevice.");
	error(RtError::DEBUG_WARNING);
	close(device_fd);
    
	return FAILURE;
    }
    
    err = ioctl(device_fd, AUDIO_GETINFO, &audio_info);
    if (err != 0) {
	sprintf(message_, "RtApiSolaris: failed to retrieve parameters in probeOpenDevice.");
	error(RtError::DEBUG_WARNING);
	close(device_fd);
    
	return FAILURE;
    }
    
    // Note: man audio(7I) specifies audio_info.play.buffer_size is unsupported
    // testcase: use record buffer
    *bufferSize = (audio_info.record.buffer_size / channels) / formatBytes(stream_.deviceFormat[mode]);

    stream_.bufferSize = *bufferSize;
    stream_.nBuffers = 1;
    
    // TODO: Verify actual device channels
    stream_.nUserChannels[mode] = channels;
    stream_.nDeviceChannels[mode] = channels;
    
    // The rest of this function is a direct copy of RtApiOss and RtApiJack
    // Set flags for buffer conversion
    stream_.doConvertBuffer[mode] = false;
    if (stream_.userFormat != stream_.deviceFormat[mode])
	stream_.doConvertBuffer[mode] = true;
    if (stream_.nUserChannels[mode] < stream_.nDeviceChannels[mode])
        stream_.doConvertBuffer[mode] = true;

    // Allocate necessary internal buffers
    if ( stream_.nUserChannels[0] != stream_.nUserChannels[1] ) {
	
	long buffer_bytes;
	if (stream_.nUserChannels[0] >= stream_.nUserChannels[1])
	    buffer_bytes = stream_.nUserChannels[0];
	else
	    buffer_bytes = stream_.nUserChannels[1];
	
	buffer_bytes *= *bufferSize * formatBytes(stream_.userFormat);
	if (stream_.userBuffer) free(stream_.userBuffer);
	stream_.userBuffer = (char *) calloc(buffer_bytes, 1);
	if (stream_.userBuffer == NULL) {
	    close(device_fd);
	    sprintf(message_, "RtApiSolaris: error allocating user buffer memory.");
	    error(RtError::DEBUG_WARNING);
	    return FAILURE;
	}
    }

    if ( stream_.doConvertBuffer[mode] ) {
	
	long buffer_bytes;
	bool makeBuffer = true;
	if ( mode == OUTPUT )
	    buffer_bytes = stream_.nDeviceChannels[0] * formatBytes(stream_.deviceFormat[0]);
	else { // mode == INPUT
	    buffer_bytes = stream_.nDeviceChannels[1] * formatBytes(stream_.deviceFormat[1]);
	    if ( stream_.mode == OUTPUT && stream_.deviceBuffer ) {
		long bytes_out = stream_.nDeviceChannels[0] * formatBytes(stream_.deviceFormat[0]);
		if ( buffer_bytes < bytes_out ) makeBuffer = false;
	    }
	}
	
	if ( makeBuffer ) {
	    buffer_bytes *= *bufferSize;
	    if (stream_.deviceBuffer) free(stream_.deviceBuffer);
	    stream_.deviceBuffer = (char *) calloc(buffer_bytes, 1);
	    if (stream_.deviceBuffer == NULL) {
	        close(device_fd);
	        sprintf(message_, "RtApiSolaris: error allocating device buffer memory.");
		error(RtError::DEBUG_WARNING);
		return FAILURE;
	    }
	}
    }
    
    stream_.device[mode] = device;
    stream_.state = STREAM_STOPPED;
    stream_.mode = mode;
    
    // Setup the buffer conversion information structure.
    if ( stream_.doConvertBuffer[mode] ) {
	if (mode == INPUT) { // convert device to user buffer
    	    stream_.convertInfo[mode].inJump = stream_.nDeviceChannels[1];
	    stream_.convertInfo[mode].outJump = stream_.nUserChannels[1];
	    stream_.convertInfo[mode].inFormat = stream_.deviceFormat[1];
	    stream_.convertInfo[mode].outFormat = stream_.userFormat;
	} else { // convert user to device buffer
	    stream_.convertInfo[mode].inJump = stream_.nUserChannels[0];
	    stream_.convertInfo[mode].outJump = stream_.nDeviceChannels[0];
	    stream_.convertInfo[mode].inFormat = stream_.userFormat;
	    stream_.convertInfo[mode].outFormat = stream_.deviceFormat[0];
	}

	if ( stream_.convertInfo[mode].inJump < stream_.convertInfo[mode].outJump )
	    stream_.convertInfo[mode].channels = stream_.convertInfo[mode].inJump;
	else
	    stream_.convertInfo[mode].channels = stream_.convertInfo[mode].outJump;
    }
    
    // interleave offset
    for (int k=0; k<stream_.convertInfo[mode].channels; k++) {
	stream_.convertInfo[mode].inOffset.push_back( k );
	stream_.convertInfo[mode].outOffset.push_back( k );
    }

    if ( stream_.apiHandle == 0 ) {
      handle = (int *) calloc(2, sizeof(int));
      stream_.apiHandle = (void *) handle;
      handle[0] = 0;
      handle[1] = 0;
    } else {
      handle = (int *) stream_.apiHandle;
    }
    handle[mode] = device_fd;
    
    return SUCCESS;
}

void RtApiSolaris :: closeStream()
{
  // We don't want an exception to be thrown here because this
  // function is called by our class destructor.  So, do our own
  // stream check.
  if ( stream_.mode == UNINITIALIZED ) {
    sprintf(message_, "RtApiSolaris::closeStream(): no open stream to close!");
    error(RtError::WARNING);
    return;
  }

  int *handle = (int *) stream_.apiHandle;

  if (stream_.callbackInfo.usingCallback) {
    stream_.callbackInfo.usingCallback = false;
    pthread_join(stream_.callbackInfo.thread, NULL);
  }

  if (handle) {
    if (handle[0]) close(handle[0]);
    if (handle[1]) close(handle[1]);
    free(handle);
    stream_.apiHandle = 0;
  }

  if (stream_.userBuffer) {
    free(stream_.userBuffer);
    stream_.userBuffer = 0;
  }

  if (stream_.deviceBuffer) {
    free(stream_.deviceBuffer);
    stream_.deviceBuffer = 0;
  }

  stream_.mode = UNINITIALIZED;
}

void RtApiSolaris :: startStream()
{
  verifyStream();
  if (stream_.state == STREAM_RUNNING) return;

  MUTEX_LOCK(&stream_.mutex);

  stream_.state = STREAM_RUNNING;

  // No need to do anything else here ... OSS automatically starts
  // when fed samples.

  MUTEX_UNLOCK(&stream_.mutex);
}

void RtApiSolaris :: stopStream()
{
  verifyStream();
  if (stream_.state == STREAM_STOPPED) return;

  // Change the state before the lock to improve shutdown response
  // when using a callback.
  stream_.state = STREAM_STOPPED;
}

void RtApiSolaris :: abortStream()
{
  stopStream();
}

void RtApiSolaris :: tickStream()
{
  verifyStream();

  int stopStream = 0;
  if (stream_.state == STREAM_STOPPED) {
    if (stream_.callbackInfo.usingCallback) usleep(50000); // sleep 50 milliseconds
    return;
  }
  else if (stream_.callbackInfo.usingCallback) {
    RtAudioCallback callback = (RtAudioCallback) stream_.callbackInfo.callback;
    stopStream = callback(stream_.userBuffer, stream_.bufferSize, stream_.callbackInfo.userData);
  }

  MUTEX_LOCK(&stream_.mutex);

  // The state might change while waiting on a mutex.
  if (stream_.state == STREAM_STOPPED)
    goto unlock;

  int result, *handle;
  char *buffer;
  int samples;
  RtAudioFormat format;
  handle = (int *) stream_.apiHandle;
  if (stream_.mode == OUTPUT || stream_.mode == DUPLEX) {

    // Setup parameters and do buffer conversion if necessary.
    if (stream_.doConvertBuffer[0]) {
      buffer = stream_.deviceBuffer;
      convertBuffer( buffer, stream_.userBuffer, stream_.convertInfo[0] );
      samples = stream_.bufferSize * stream_.nDeviceChannels[0];
      format = stream_.deviceFormat[0];
    }
    else {
      buffer = stream_.userBuffer;
      samples = stream_.bufferSize * stream_.nUserChannels[0];
      format = stream_.userFormat;
    }

    // Do byte swapping if necessary.
    if (stream_.doByteSwap[0])
      byteSwapBuffer(buffer, samples, format);

    // Write samples to device.
    result = write(handle[0], buffer, samples * formatBytes(format));

    if (result == -1) {
      // This could be an underrun, but the basic Solaris API doesn't provide a means for determining that.
      sprintf(message_, "RtApiSolaris: audio write error for device (%s).",
              devices_[stream_.device[0]].name.c_str());
      error(RtError::DRIVER_ERROR);
    }
  }

  if (stream_.mode == INPUT || stream_.mode == DUPLEX) {

    // Setup parameters.
    if (stream_.doConvertBuffer[1]) {
      buffer = stream_.deviceBuffer;
      samples = stream_.bufferSize * stream_.nDeviceChannels[1];
      format = stream_.deviceFormat[1];
    }
    else {
      buffer = stream_.userBuffer;
      samples = stream_.bufferSize * stream_.nUserChannels[1];
      format = stream_.userFormat;
    }

    // Read samples from device.
    result = read(handle[1], buffer, samples * formatBytes(format));

    if (result == -1) {
      // This could be an overrun, but the basic Solaris API doesn't provide a means for determining that.
      sprintf(message_, "RtApiSolaris: audio read error for device (%s).",
              devices_[stream_.device[1]].name.c_str());
      error(RtError::DRIVER_ERROR);
    }

    // Do byte swapping if necessary.
    if (stream_.doByteSwap[1])
      byteSwapBuffer(buffer, samples, format);

    // Do buffer conversion if necessary.
    if (stream_.doConvertBuffer[1])
      convertBuffer( stream_.userBuffer, stream_.deviceBuffer, stream_.convertInfo[1] );
  }

 unlock:
  MUTEX_UNLOCK(&stream_.mutex);

  if (stream_.callbackInfo.usingCallback && stopStream)
    this->stopStream();

}

void RtApiSolaris :: setStreamCallback(RtAudioCallback callback, void *userData)
{
  verifyStream();

  CallbackInfo *info = (CallbackInfo *) &stream_.callbackInfo;
  if ( info->usingCallback ) {
    sprintf(message_, "RtApiSolaris: A callback is already set for this stream!");
    error(RtError::WARNING);
    return;
  }

  info->callback = (void *) callback;
  info->userData = userData;
  info->usingCallback = true;
  info->object = (void *) this;

  // Set the thread attributes for joinable and realtime scheduling
  // priority.  The higher priority will only take affect if the
  // program is run as root or suid.
  pthread_attr_t attr;
  pthread_attr_init(&attr);
  pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
  pthread_attr_setschedpolicy(&attr, SCHED_RR);

  int err = pthread_create(&(info->thread), &attr, solarisCallbackHandler, &stream_.callbackInfo);
  pthread_attr_destroy(&attr);
  if (err) {
    info->usingCallback = false;
    sprintf(message_, "RtApiSolaris: error starting callback thread!");
    error(RtError::THREAD_ERROR);
  }
}

void RtApiSolaris :: cancelStreamCallback()
{
  verifyStream();

  if (stream_.callbackInfo.usingCallback) {

    if (stream_.state == STREAM_RUNNING)
      stopStream();

    MUTEX_LOCK(&stream_.mutex);

    stream_.callbackInfo.usingCallback = false;
    pthread_join(stream_.callbackInfo.thread, NULL);
    stream_.callbackInfo.thread = 0;
    stream_.callbackInfo.callback = NULL;
    stream_.callbackInfo.userData = NULL;

    MUTEX_UNLOCK(&stream_.mutex);
  }
}

extern "C" void *solarisCallbackHandler(void *ptr)
{
  CallbackInfo *info = (CallbackInfo *) ptr;
  RtApiSolaris *object = (RtApiSolaris *) info->object;
  bool *usingCallback = &info->usingCallback;

  while ( *usingCallback ) {
    pthread_testcancel();
    try {
      object->tickStream();
    }
    catch (RtError &exception) {
      fprintf(stderr, "\nRtApiSolaris: callback thread error (%s) ... closing thread.\n\n",
              exception.getMessageString());
      break;
    }
  }

  return 0;
}

//******************** End of __SOLARIS_AUDIO__ *********************//
#endif

